module m_age_algorithms
    implicit none
    save

! Written by Willi Kappler
! willi.kappler@uni-tuebingen.de

    !> Make everything private
    private

    !> Universal constants
    real(8), parameter :: PI = 3.1415926535897932384626433
    real(8), parameter :: univ_gas_const = 8.314462175
    real(8), parameter :: days_per_year = 365.25
    real(8), parameter :: hours_per_day = 24.0
    real(8), parameter :: sec_per_hour = 3600.0
    real(8), parameter :: mil_year = 1.0e6
    real(8), parameter :: sec_per_mil_year = sec_per_hour * hours_per_day * days_per_year * mil_year
    real(8), parameter :: kelvin_to_deg_cel = -273.15
    real(8), parameter :: deg_cel_to_kelvin = +273.15
    real(8), parameter :: um_to_m = 1.0e-6
    real(8), parameter :: kJ_to_J = 1.0e3

    character(30) :: method

    !> Exported subroutines
    public age_with_closure
    public Muscovite
    public ZFT
    public Apatite_U_Th_Pb
    public Biotite
    public Rutile_U_Pb
    public Titanite_U_Pb
    public Zircon_U_Pb

    contains

        !> Calculates the muscovite age given the time / temperature history path
        !! @param temperature Array of temperature values temperature path [degree celsius]
        !! @param time Array of time values / time path [My]
        !! @param n_elems Number of elements for the above arrays
        !! @param age Output parameter for the calculated age value
        subroutine Muscovite(temperature, time, n_elems, age, m, i)
            implicit none

            !> Subroutine parameters
            integer(4), intent(in) :: n_elems, m, i

            real(8), intent(in), dimension(n_elems) :: temperature, time

            real(8), intent(out) :: age

            !> Local variables
            real(8) :: age_real8

            !> local constants
            real(8), parameter :: D0 = 4.0e-8_8
            real(8), parameter :: energy = 180.0_8
            real(8), parameter :: grain_size = 750.0_8
            real(8), parameter :: geometry_factor = 27.0_8 ! cylinder geometry, changes from Byron Adams

            method = "Muscovite Ar"
            call age_with_closure(dble(temperature), dble(time), n_elems, D0, energy, grain_size, geometry_factor, age_real8, m, i)

            age = real(age_real8)
        end subroutine Muscovite

        !> Calculates the zircon fission track age given the time / temperature history path
        !! @param temperature Array of temperature values temperature path [degree celsius]
        !! @param time Array of time values / time path [My]
        !! @param n_elems Number of elements for the above arrays
        !! @param age Output parameter for the calculated age value
        subroutine ZFT(temperature, time, n_elems, age, m, i)
            implicit none

            !> Subroutine parameters
            integer(4), intent(in) :: n_elems, m, i

            real(8), intent(in), dimension(n_elems) :: temperature, time

            real(8), intent(out) :: age

            !> Local variables
            real(8) :: age_real8

            !> local constants
            real(8), parameter :: D0 = 0.001_8
            real(8), parameter :: energy = 208.32768_8
            real(8), parameter :: grain_size = 3.158_8
            real(8), parameter :: geometry_factor = 55.0_8

            method = "ZFT"
            call age_with_closure(dble(temperature), dble(time), n_elems, D0, energy, grain_size, geometry_factor, age_real8, m, i)

            age = real(age_real8)
        end subroutine ZFT

        !> Calculates the apatite U-Th/Pb age given the time / temperature history path
        !! @param temperature Array of temperature values temperature path [degree celsius]
        !! @param time Array of time values / time path [My]
        !! @param n_elems Number of elements for the above arrays
        !! @param age Output parameter for the calculated age value
        subroutine Apatite_U_Th_Pb(temperature, time, n_elems, age, m, i)
            implicit none

            !> Subroutine parameters
            integer(4), intent(in) :: n_elems, m, i

            real(8), intent(in), dimension(n_elems) :: temperature, time

            real(8), intent(out) :: age

            !> Local variables
            real(8) :: age_real8

            !> local constants
            real(8), parameter :: D0 = 2.0e-8_8
            real(8), parameter :: energy = 230.0_8
            real(8), parameter :: grain_size = 50.0_8
            real(8), parameter :: geometry_factor = 55.0_8

            method = "Apatite U-Th / Pb (APb)"
            call age_with_closure(dble(temperature), dble(time), n_elems, D0, energy, grain_size, geometry_factor, age_real8, m, i)

            age = real(age_real8)
        end subroutine Apatite_U_Th_Pb

        !> Calculates the Rutile U-Pb (RUPb) age given the time / temperature history path
        !! @param temperature Array of temperature values temperature path [degree celsius]
        !! @param time Array of time values / time path [My]
        !! @param n_elems Number of elements for the above arrays
        !! @param age Output parameter for the calculated age value
        subroutine Rutile_U_Pb(temperature, time, n_elems, age, m, i)
            implicit none

            !> Subroutine parameters
            integer(4), intent(in) :: n_elems, m, i

            real(8), intent(in), dimension(n_elems) :: temperature, time

            real(8), intent(out) :: age

            !> Local variables
            real(8) :: age_real8

            !> local constants
            real(8), parameter :: D0 = 1.6e-10_8
            real(8), parameter :: energy = 243.0_8
            real(8), parameter :: grain_size = 250.0_8
            real(8), parameter :: geometry_factor = 55.0_8 ! spherical

            method = "Rutile U-Pb (RUPb)"
            call age_with_closure(dble(temperature), dble(time), n_elems, D0, energy, grain_size, geometry_factor, age_real8, m, i)

            age = real(age_real8)
        end subroutine Rutile_U_Pb

        !> Calculates the Titanite U-Pb (TUPb) age given the time / temperature history path
        !! @param temperature Array of temperature values temperature path [degree celsius]
        !! @param time Array of time values / time path [My]
        !! @param n_elems Number of elements for the above arrays
        !! @param age Output parameter for the calculated age value
        subroutine Titanite_U_Pb(temperature, time, n_elems, age, m, i)
            implicit none

            !> Subroutine parameters
            integer(4), intent(in) :: n_elems, m, i

            real(8), intent(in), dimension(n_elems) :: temperature, time

            real(8), intent(out) :: age

            !> Local variables
            real(8) :: age_real8

            !> local constants
            real(8), parameter :: D0 = 1.1e-4_8
            real(8), parameter :: energy = 331.0_8
            real(8), parameter :: grain_size = 500.0_8
            real(8), parameter :: geometry_factor = 55.0_8 ! spherical

            method = "Titanite U-Pb (TUPb)"
            call age_with_closure(dble(temperature), dble(time), n_elems, D0, energy, grain_size, geometry_factor, age_real8, m, i)

            age = real(age_real8)
        end subroutine Titanite_U_Pb

        !> Calculates the Zircon U-Pb (TUPb) age given the time / temperature history path
        !! @param temperature Array of temperature values temperature path [degree celsius]
        !! @param time Array of time values / time path [My]
        !! @param n_elems Number of elements for the above arrays
        !! @param age Output parameter for the calculated age value
        subroutine Zircon_U_Pb(temperature, time, n_elems, age, m, i)
            implicit none

            !> Subroutine parameters
            integer(4), intent(in) :: n_elems, m, i

            real(8), intent(in), dimension(n_elems) :: temperature, time

            real(8), intent(out) :: age

            !> Local variables
            real(8) :: age_real8

            !> local constants
            real(8), parameter :: D0 = 7.8e-3_8
            real(8), parameter :: energy = 544.0_8
            real(8), parameter :: grain_size = 50.0_8
            real(8), parameter :: geometry_factor = 55.0_8 ! spherical

            method = "Zircon U-Pb (ZUPb)"
            call age_with_closure(dble(temperature), dble(time), n_elems, D0, energy, grain_size, geometry_factor, age_real8, m, i)

            age = real(age_real8)
        end subroutine Zircon_U_Pb

        !> Calculates the biotite age given the time / temperature history path
        !! @param temperature Array of temperature values temperature path [degree celsius]
        !! @param time Array of time values / time path [My]
        !! @param n_elems Number of elements for the above arrays
        !! @param age Output parameter for the calculated age value
        subroutine Biotite(temperature, time, n_elems, age, m, i)
            implicit none

            !> Subroutine parameters
            integer(4), intent(in) :: n_elems, m, i

            real(8), intent(in), dimension(n_elems) :: temperature, time

            real(8), intent(out) :: age

            !> Local variables
            real(8) :: age_real8

            !> local constants
            real(8), parameter :: D0 = 2.0e-13_8
            real(8), parameter :: energy = 105.0_8
            real(8), parameter :: grain_size = 500.0_8
            real(8), parameter :: geometry_factor = 27.0_8 ! cylinder geometry, changes from Byron Adams

            method = "Biotite"
            call age_with_closure(dble(temperature), dble(time), n_elems, D0, energy, grain_size, geometry_factor, age_real8, m, i)

            age = real(age_real8)
        end subroutine Biotite

        !> Calculates the age using the closure temperature
        !! @param temperature Array of temperature values temperature path [degree celsius]
        !! @param time Array of time values / time path [My]
        !! @param n_elems Number of elements for the above arrays
        !! @param D0 Diffusivity at infinite temperature [m² / s]
        !! @param energy Activation energy [kJ / mol]
        !! @param grain_size Grain size [µm]
        !! @param geometry_factor Geometry factor for plane sheets
        !! @param age Output parameter for the calculated age value
        subroutine age_with_closure(temperature, time, n_elems, D0, energy, grain_size, geometry_factor, age, m, i)
            implicit none

            !> Subroutine parameters
            integer(4), intent(in) :: n_elems, m, i

            real(8), intent(in) :: D0, energy, grain_size, geometry_factor

            real(8), intent(in), dimension(n_elems) :: temperature, time

            real(8), intent(out) :: age

            !> local variables
            real(8), dimension(n_elems) :: local_time
            real(8) :: cooling_rate !> [1 deg / 10 My]
            real(8) :: closure_temp !> [degree celsius]
            real(8) :: prev_closure_temp !> [degree celsius]
            real(8) :: prev_temp !> [degree celsius]
            real(8) :: tau
            real(8) :: diff
            real(8) :: ratio
            real(8) :: energy_in_J !> Activation energy in [J]

            integer(4) :: j

            !> What is diff ?
            diff = (D0 / (grain_size * um_to_m)**2.0) * sec_per_mil_year

            !> Adjust time to local time
            local_time = time - time(n_elems)

            !> Initial age value
            age = local_time(1)

            !> Convert activation energy from kJ to J
            energy_in_J = energy * kJ_to_J

            !> Initial value for variables
            cooling_rate = 0.0
            closure_temp = 0.0
            prev_closure_temp = 0.0
            prev_temp = 0.0
            tau = 0.0
            ratio = 0.0

            !> Iterative age calculation
            do j = n_elems, 2, -1
                if (j == 1) then
                    cooling_rate = (temperature(j + 1) - temperature(j)) / (local_time(j + 1) - local_time(j))
                else if (j == n_elems) then
                    cooling_rate = (temperature(j) - temperature(j - 1)) / (local_time(j) - local_time(j - 1))
                else
                    cooling_rate = (temperature(j + 1) - temperature(j - 1)) / (local_time(j + 1) - local_time(j - 1))
                end if

                !> Cooling rate must be a valid value
                cooling_rate = max(cooling_rate, 0.1_8)

                !> What is tau ?
                tau = univ_gas_const * ((temperature(j) + deg_cel_to_kelvin)**2.0) / (energy_in_J * cooling_rate)
                closure_temp = kelvin_to_deg_cel + (energy_in_J / (log(geometry_factor * tau * diff) * univ_gas_const))

!                print *, "j: ", j, ", cooling: ", cooling_rate, ", tau: ", tau, ", closure: ", closure_temp
!                print *, "diff: ", diff
!                print *, "temperature(j): ", temperature(j), ", local_time(j): ", local_time(j)
!                print *, "temperature(j-1): ", temperature(j-1), ", local_time(j-1): ", local_time(j-1)

                if (temperature(j) > closure_temp) then
                    ratio = (prev_closure_temp - prev_temp) / (prev_closure_temp - prev_temp + temperature(j) - closure_temp)
                    age = local_time(j) + (ratio * (local_time(j-1) - local_time(j)))
                    return
                end if

                prev_closure_temp = closure_temp
                prev_temp = temperature(j)

                write (80, *) m, i, method, prev_temp, prev_closure_temp

            end do
        end subroutine age_with_closure
end module m_age_algorithms

